proc NCBI_Blast_init {} {
    #Creating menu buttons
    .menubar.module.menu add command -label "NCBI-Blast" -command "NCBI_Blast"
    .menubar.module.menu add command -label "NCBI-References" -command "NCBI_Blast_ref"
}

proc NCBI_Blast_remove {} {
    #Remove menu buttons
    .menubar.module.menu delete "NCBI-Blast"
    .menubar.module.menu delete "NCBI-References"
}

proc NCBI_Blast {} {
    global env
    global ncbi_blast_pb
    global ncbi_blast_sb
    global ncbi_blast_expectCheck
    variable ncbi_blast_evalue
    variable ncbi_blast_hit
    variable ncbi_blast_threads
    variable ncbi_blast_organism
    
    if {$env(dbName) ne " "} {
        set ncbi_blast_expectCheck 1
        set ncbi_blast_evalue "1e-003"
        set ncbi_blast_hit 5
        set ncbi_blast_threads 5
        
        #  Configure the main window
        set t .ncbi_blast
        catch {destroy $t}
        toplevel $t
        wm title $t "NCBI Blast - $env(dbName)"
        wm minsize $t 250 170
        wm resizable $t 1 0
        
        frame $t.blast -bd 2 -relief groove
        label $t.blast.label -text "Run blast for:"
        ComboBox $t.blast.cb -width 18 -hottrack 1 -background white -text "selected sequences" -values "{selected sequences} {all (without db hit)} [.result.notebook1.nb.fdbb.0.feat cget -values]"
        
        frame $t.program
        frame $t.program.p
        label $t.program.p.clabel -text "Program:"
        ComboBox $t.program.p.cb -width 15 -hottrack 1 -background white -text "tblastx" -values "megablast blastn blastp rpsblast blastx tblastn tblastx"
        frame $t.database
        frame $t.database.p
        label $t.database.p.clabel -text "Database:"
        ComboBox $t.database.p.cb -width 15 -hottrack 1 -background white -text "nt" -values "nt nr refseq_rna refseq_genomic chromosome est est_others gss pat pdb"
        frame $t.organism
        frame $t.organism.p
        label $t.organism.p.clabel -text "Organism/TaxID:"
        ComboBox $t.organism.p.cb -width 15 -hottrack 1 -background white -text "" -values "{} {Homo sapiens} {Mus Mucullus}"
        
        frame $t.hitnum
        label $t.hitnum.label -text "Max. number of hits:"
        entry $t.hitnum.ent -textvariable ncbi_blast_hit -width 5
        frame $t.evalue
        checkbutton $t.evalue.check -text "Max. Expect Value:" -variable ncbi_blast_expectCheck
        entry $t.evalue.ent -textvariable ncbi_blast_evalue -width 7
        frame $t.threadnum
        label $t.threadnum.label -text "Number of NCBI requests:"
        entry $t.threadnum.ent -textvariable ncbi_blast_threads -width 5
        
        pack $t.blast.label -side left; pack $t.blast.cb -side right
        pack $t.program.p.clabel -side left; pack $t.program.p.cb -side right
        pack $t.database.p.clabel -side left; pack $t.database.p.cb -side right
        pack $t.organism.p.clabel -side left; pack $t.organism.p.cb -side right
        pack $t.program.p -side top -fill both
        pack $t.database.p -side top -fill both
        pack $t.organism.p -side top -fill both
        pack $t.hitnum.label -side left; pack $t.hitnum.ent -side right
        pack $t.evalue.check -side left; pack $t.evalue.ent -side right
        pack $t.threadnum.label -side left; pack $t.threadnum.ent -side right
        pack $t.blast $t.program $t.database $t.organism $t.hitnum $t.evalue $t.threadnum -side top -fill both
        
        #The Buttons Close and Help
        frame $t.buts -bd 2 -relief groove
        button $t.buts.ok -text "Run" -command "ncbi_blast_search $env(dbName) blast"
        button $t.buts.closeW -text "Close Window" -command "destroy $t"
        pack $t.buts.ok $t.buts.closeW -side left -expand 1 -anchor c
        pack $t.buts -side bottom -fill both
        
        wm geometry $t +[expr [winfo rootx .]+[winfo width .]/2]+[expr [winfo rooty .]+[winfo height .]/2]
        raise $t
    } else {
        ::errorPopUp "Error" "No database selected."
    }
}

proc NCBI_Blast_ref {} {
    global env
    global ncbi_blast_expectCheck
    variable ncbi_blast_threads
    
    if {$env(dbName) ne " "} {        
        #  Configure the main window
        set ncbi_blast_expectCheck 1
        set ncbi_blast_threads 5
        
        set t .ncbi_blast_ref
        toplevel $t
        wm title $t "NCBI Blast - get reference"
        wm minsize $t 250 170
        wm resizable $t 1 0
        
        frame $t.blast -bd 2 -relief groove
        label $t.blast.label -text "Get references of:"
        ComboBox $t.blast.cb -width 18 -hottrack 1 -background white -text "selected sequences" -values "{selected sequences} {all (without db reference)}"
        
        frame $t.database
        frame $t.database.p
        label $t.database.p.clabel -text "Database:"
        ComboBox $t.database.p.cb -width 15 -hottrack 1 -background white -text "nt" -values "nt nr refseq_rna refseq_genomic chromosome est est_others gss pat pdb"
        
        pack $t.blast.label -side left; pack $t.blast.cb -side right
        pack $t.database.p.clabel -side left; pack $t.database.p.cb -side right
        pack $t.database.p -side top -fill both
        
        frame $t.threadnum
        label $t.threadnum.label -text "Number of NCBI requests:"
        entry $t.threadnum.ent -textvariable ncbi_blast_threads -width 5
        
        pack $t.threadnum.label -side left; pack $t.threadnum.ent -side right
        pack $t.blast $t.database $t.threadnum -side top -fill both
        
        #The Buttons Close and Help
        frame $t.buts -bd 2 -relief groove
        button $t.buts.ok -text "Run" -command "ncbi_blast_search $env(dbName) ref"
        button $t.buts.closeW -text "Close Window" -command "destroy $t"
        pack $t.buts.ok $t.buts.closeW -side left -expand 1 -anchor c
        pack $t.buts -side bottom -fill both
        
        wm geometry $t +[expr [winfo rootx .]+[winfo width .]/2]+[expr [winfo rooty .]+[winfo height .]/2]
        raise $t
    } else {
        ::errorPopUp "Error" "No database selected."
    }
}

proc ncbi_blast_search {dbName mode} {
    global env
    global ncbi_blast_expectCheck
    global ncbi_thread_num
    global ncbi_finished_threads
    global evalue
    global organism
    global thread_resultsArr
    
    set ncbi_thread_num 0
    set ncbi_finished_threads 0
    set id_seqL ""
    
    if {$mode eq "blast"} {
        if {[.ncbi_blast.blast.cb get] eq "selected sequences"} {
            set page [.result.notebook1.nb raise]
            if {$page ne "sb" && $page ne "db_stat" && $page ne "process" && $page ne "blast" && $page ne "db_info"} {
                set t .result.notebook1.nb.f$page.1.table.l
                $t finishediting
                set stop 1
                set i 0
                while {$stop} {
                    if {[lindex [$t get $i] 0] eq 1} {
                        lappend id_seqL "[lindex [$t get $i] 23] [lindex [$t get $i] 24]"
                    } elseif {[lindex [$t get $i] 0] eq ""} {
                        set stop 0
                    }
                    incr i
                }
            } else {
                ::errorPopUp "Error" "Please select the 'Database Browser' or a project tab."
            }
        } elseif {[.ncbi_blast.blast.cb get] eq "all (without db hit)"} {
            set tmpL [::get_seq_info "blast_noDB [.ncbi_blast.database.p.cb get]" id_seq]
            foreach id_seq $tmpL {
                lappend id_seqL "$id_seq $dbName"
            }
            unset tmpL
        } else {
            set tmpL [::get_seq_info "[.ncbi_blast.blast.cb get]" id_seq]
            foreach id_seq $tmpL {
                lappend id_seqL "$id_seq $dbName"
            }
            unset tmpL
        }
        set database [.ncbi_blast.database.p.cb get]
        set program [.ncbi_blast.program.p.cb get]
        set organism [.ncbi_blast.organism.p.cb get]
        set evalue [.ncbi_blast.evalue.ent get]
        set threadNum [.ncbi_blast.threadnum.ent get]
        set hit [.ncbi_blast.hitnum.ent get]
        set url "http://www.ncbi.nlm.nih.gov/"  
        catch {destroy .ncbi_blast}
    } elseif {$mode eq "ref"} {
        set database [.ncbi_blast_ref.database.p.cb get]
        set program -
        set organism -
        set evalue -
        set threadNum [.ncbi_blast_ref.threadnum.ent get]
        set hit -
        set url "http://www.ncbi.nlm.nih.gov/"
        if {[.ncbi_blast_ref.blast.cb get] eq "selected sequences"} {
            set page [.result.notebook1.nb raise]
            if {$page ne "sb" && $page ne "db_stat" && $page ne "process" && $page ne "blast" && $page ne "db_info"} {
                set t .result.notebook1.nb.f$page.1.table.l
                $t finishediting
                set stop 1
                set i 0
                set tmpString ""
                while {$stop} {
                    if {[lindex [$t get $i] 0] eq 1} {
                        append tmpString ", '[lindex [$t get $i] 23]'"
                    } elseif {[lindex [$t get $i] 0] eq ""} {
                        set stop 0
                    }
                    incr i
                }
                set tmpString [string range $tmpString 2 end]
                set sql_res [eval [subst {thread::send $env(phpThreadID) {::php_getSQLresults "SELECT DISTINCT id FROM $env(dbName).blast_hits WHERE db = '$database' AND id_seq IN ($tmpString) AND id NOT IN (SELECT id FROM refdb.$database);"}}]]
                foreach sRes $sql_res {
                    lappend id_seqL [lindex $sRes 0]
                }
                unset tmpString
            } else {
                ::errorPopUp "Error" "Please select the 'Database Browser' or a project tab."
            }
        } elseif {[.ncbi_blast_ref.blast.cb get] eq "all (without db reference)"} {
            set sql_res [eval [subst {thread::send $env(phpThreadID) {::php_getSQLresults "SELECT DISTINCT id FROM $env(dbName).blast_hits WHERE db = '$database' AND id NOT IN (SELECT id FROM refdb.$database);"}}]]
            foreach sRes $sql_res {
                lappend id_seqL [lindex $sRes 0]
            }
        }
        catch {destroy .ncbi_blast_ref}
    }
    
    if {$id_seqL ne ""} {
        catch {
            eval [subst {thread::send $::threadPHP {::php_getSQLresults "CREATE TABLE refdb.$database (id text NOT NULL, id2 text, organism text, description text, product text, CONSTRAINT ${database}_pkey PRIMARY KEY (id));"}}]
            eval [subst {thread::send $::threadPHP {::php_getSQLresults "CREATE INDEX index_id_$database ON refdb.$database USING btree (id);"}}]
            eval [subst {thread::send $::threadPHP {::php_getSQLresults "CREATE INDEX index_organism_$database ON refdb.$database USING btree (organism);"}}]
            eval [subst {thread::send $::threadPHP {::php_getSQLresults "CREATE INDEX index_description_$database ON refdb.$database USING btree (description);"}}]
            eval [subst {thread::send $::threadPHP {::php_getSQLresults "CREATE INDEX index_product_$database ON refdb.$database USING btree (product);"}}] 
        }
        
        set processID [llength $env(processL)]
        set env($processID.ncbi_blast_num) 0
        
        set chunkSize [expr [llength $id_seqL] / $threadNum]
        if {$chunkSize < 1} {
            set chunkSize 1
        }
        for {set i 0} {$i < $threadNum} {incr i} {
            set thread_resultsArr($i) "Thread $i: 0 of N/A processed; 0 with hits; 0 without hits; 0 with error"
            set threadArrID($i) [thread::create {
                package require http
                
                proc ncbiblast {homeFolder db program expect hit org url id_seqL id_seqL_num expectCheck mode processID sequence_type mainThreadID phpTID tNum} {
                    global env
                    global phpThreadID 
                    global ncbi_blast_expectCheck
                    global evalue
                    global hitNum
                    global organism
                    global database
                    global threadNum
                    global queHitNum 
                    global queNoHitNum
                    global queErrorNum
                    global queTotalNum
                    
                    set phpThreadID $phpTID
                    set ncbi_blast_expectCheck $expectCheck
                    set evalue $expect
                    set hitNum $hit
                    set organism $org
                    set database $db
                    set threadNum $tNum
                    set queHitNum 0
                    set queNoHitNum 0
                    set queErrorNum 0
                    set queTotalNum [llength $id_seqL]
                    
                    if {$mode eq "blast"} {
                        if {$program eq "megablast"} {
                            set program "blastn&MEGABLAST=on"
                        } elseif {$program eq "rpsblast"} {
                            set program "blastp&SERVICE=rpsblast"
                        }
                        foreach id_seq $id_seqL {
                            if {$sequence_type eq "nucleotide"} {
                                set sql_res [eval [subst {thread::send $phpThreadID {::php_getSQLresults "SELECT name, sequence FROM [lindex $id_seq 1].sequences_nucleotide WHERE id_seq = '[lindex $id_seq 0]'"}}]]
                            } else {
                                set sql_res [eval [subst {thread::send $phpThreadID {::php_getSQLresults "SELECT name, sequence FROM [lindex $id_seq 1].sequences_protein WHERE id_prot = '[lindex $id_seq 0]'"}}]]
                            }
                            foreach sRes $sql_res {
                                set query [lindex $sRes 0]
                                set sequence [lindex $sRes 1]
                            }
                            unset sql_res
                            set que [::http::formatQuery >$query\n$sequence]
                            if {$organism ne ""} {
                                set cmd "CMD=Put&PROGRAM=$program&DATABASE=$database&qorganism=$organism&QUERY=$que"
                            } else  {
                                set cmd "CMD=Put&PROGRAM=$program&DATABASE=$database&QUERY=$que"
                            }
                            set str ""
                            set i 0
                            while {$i < 10} {
                                if {[catch {set token [::http::geturl $url/blast/Blast.cgi -query $cmd -timeout 25000]}]} {
                                    sleep 5000
                                    incr i
                                } else {
                                    set str [::http::data $token]
                                    set i 10
                                }
                            }
                            catch {http::cleanup $token}
                            if {$str ne ""} {
                                set rtoe ""
                                set rid ""
                                set str [regsub -all {"} $str ""]
                                set str "{[regsub -all {\n} $str "\} \{"]}"
                                foreach line $str {
                                    if {[lindex $line 0] == "RID"} {
                                        set rid [lindex $line 2]
                                    }
                                    if {[lindex $line 0] == "RTOE"} {
                                        set rtoe [lindex $line 2]
                                        break
                                    }
                                }
                                if {$rtoe ne "" && $rid ne ""} {
                                    set gotHits 0
                                    sleep [expr $rtoe * 1000]
                                    set complete 0
                                    while {$complete eq 0} {
                                        if {[catch {set token [http::geturl $url/blast/Blast.cgi -query "CMD=Get&FORMAT_OBJECT=SearchInfo&RID=$rid" -timeout 25000]}]} {
                                            sleep 5000
                                        } else {
                                            set str [::http::data $token]
                                            http::cleanup $token
                                            set str [regsub -all {[^a-zA-Z ]} $str ""]
                                            for {set i 0} {$i < [llength $str]} {incr i} {
                                                if {[lindex $str $i] == "QBlastInfoBegin"} {
                                                    incr i
                                                    set status [lindex $str $i]
                                                    set results [string index [lindex $str [expr $i + 3]] 27]
                                                    unset str
                                                    break
                                                }
                                            }
                                            if {$status == "StatusREADY"} {
                                                if {$results eq "y"} {
                                                    set i 0
                                                    while {$complete eq 0} {
                                                        if {[catch {set token [http::geturl $url/blast/Blast.cgi -query "CMD=Get&FORMAT_TYPE=Text&RID=$rid" -timeout 25000]}]} {
                                                            sleep 5000
                                                            incr i
                                                        } else {
                                                            set blastRes [regsub -all {["]} [::http::data $token] ""]
                                                            http::cleanup $token
                                                            set blastRes [regsub -all {[\{]} $blastRes "("]
                                                            set blastRes [regsub -all {[\}]} $blastRes ")"]
                                                            set hitNumber [ncbi_blast_get_hits [lindex $id_seq 0] [lindex $id_seq 1] $blastRes $hitNum $url $sequence_type]
                                                            set gotHits "$hitNumber"
                                                            set complete 1
                                                        }
                                                        if {$i eq 10} {
                                                            set complete 1
                                                        }
                                                    }
                                                } else {
                                                    set complete 1
                                                }
                                            } elseif {$status == "StatusWAITING"} {
                                                sleep 5000
                                            } elseif {$status == "StatusFAILED"} {
                                                set complete 1
                                            } elseif {$status == "StatusUNKNOWN"} {
                                                set complete 1
                                            }
                                        }
                                    }
                                }
                            } else {
                                set gotHits "no_respond"
                            }
                            returnResult $processID running "$id_seqL_num $db $program" [lindex $id_seq 1] "$gotHits" $mainThreadID blast
                        }
                        returnResult $processID finished "$id_seqL_num $db $program" $env(dbName) - $mainThreadID blast
                    } elseif {$mode eq "ref"} {
                        foreach id $id_seqL {
                            set ref "[regsub -all {'} [ncbi_blast_get_refs $id $url] "''"]"
                            lappend refL "'[lindex $ref 0]', '[lindex $ref 1]', '[lindex $ref 2]', '[lindex $ref 3]', '[lindex $ref 4]'"
                            addRes - "$refL" "" -
                            unset ref
                            returnResult $processID running "$id_seqL_num" $env(dbName) "$refL" $mainThreadID ref
                            unset refL
                        }
                        returnResult $processID finished "$id_seqL_num" $env(dbName) - $mainThreadID ref
                    }
                    
                    unset id_seqL
                }
                
                proc ncbi_blast_get_hits {id_seq dbName blastRes hitNum url sequence_type} {
                    global ncbi_blast_expectCheck
                    global evalue
                    global organism
                    global database
                    
                    set refL ""
                    set hitL ""
                    set blastHitL ""
                    set hitKey ""
                    set length ""
                    set score ""
                    set expect ""
                    set perfection ""
                    set start -1
                    set end -1
                    
                    set count 1
                    set id ""
                    set mLStart ""
                    set str [regsub -all {["]} $blastRes ""]
                    set str "{[regsub -all {\n} $blastRes "\} \{"]}"
                    
                    foreach line $str {
                        if {[string index $line 0] eq ">"} {
                            if {$score ne ""} {
                                if {$count <= $hitNum} {
                                    if {$id ne "$hitID"} {
                                        set length ???
                                        catch {
                                            if {$start > $end} {
                                                set length [expr $start - $end + 1]
                                            } else {
                                                set length [expr $end - $start + 1]
                                            }
                                        }
                                        if {$sequence_type eq "nucleotide"} {
                                            lappend blastHitL "'$id_seq', '-1', {$hitID} '$database', '$start', '$end', '$length', '$expect', '$score', '$perfection' $expect"
                                        } else  {
                                            lappend blastHitL "'-1', '$id_seq', {$hitID} '$database', '$start', '$end', '$length', '$expect', '$score', '$perfection' $expect"
                                        }
                                        set id $hitID
                                        incr count
                                    }
                                    set hitID [lindex [regsub -all {[|.]} [lindex $line 0] " "] 1]
                                    set score ""
                                    set length ""
                                    set score ""
                                    set expect ""
                                    set perfection ""
                                    set start -1
                                    set end -1
                                }
                            } else {
                                set hitID [lindex [regsub -all {[|.]} [lindex $line 0] " "] 1]
                            }
                        } elseif {[lindex $line 0] eq "Score"} {
                            if {$score ne ""} {
                                if {$count <= $hitNum} {
                                    if {$id ne "$hitID"} {
                                        set length ???
                                        catch {
                                            if {$start > $end} {
                                                set length [expr $start - $end + 1]
                                            } else {
                                                set length [expr $end - $start + 1]
                                            }
                                        }
                                        if {$sequence_type eq "nucleotide"} {
                                            lappend blastHitL "'$id_seq', '-1', {$hitID} '$database', '$start', '$end', '$length', '$expect', '$score', '$perfection' $expect"
                                        } else  {
                                            lappend blastHitL "'-1', '$id_seq', {$hitID} '$database', '$start', '$end', '$length', '$expect', '$score', '$perfection' $expect"
                                        }
                                        set id $hitID
                                        incr count
                                        set score ""
                                        set length ""
                                        set score ""
                                        set expect ""
                                        set perfection ""
                                        set start -1
                                        set end -1
                                    }
                                }
                            }
                            set score [lindex $line 2]
                            set expect [lindex $line 7]
                            if {[string index $expect end] eq ","} {
                                set expect [string range $expect 0 [expr [string length $expect - 2]]]
                            }
                            if {[string index $expect 0] eq "e"} {
                                set expect "1$expect"
                            }
                        } elseif {[lindex $line 0] eq "Identities"} {
                            set perfection [regsub -all {[,()]} [lindex $line 3] ""]
                        } elseif {[lindex $line 0] eq "Query"} {
                            if {$start eq -1} {
                                set start [lindex $line 1]
                                set end [lindex $line end]
                            } else {
                                set end [lindex $line end]
                            }
                            set matchLine 1
                        }
                    }
                    unset str
                    if {$score ne "" && $count <= $hitNum} {
                        if {$id ne "$hitID"} {
                            set length ???
                            catch {
                                if {$start > $end} {
                                    set length [expr $start - $end + 1]
                                } else {
                                    set length [expr $end - $start + 1]
                                }
                            }
                            if {$sequence_type eq "nucleotide"} {
                                lappend blastHitL "'$id_seq', '-1', {$hitID} '$database', '$start', '$end', '$length', '$expect', '$score', '$perfection' $expect"
                            } else  {
                                lappend blastHitL "'-1', '$id_seq', {$hitID} '$database', '$start', '$end', '$length', '$expect', '$score', '$perfection' $expect"
                            }
                            set id $hitID
                            incr count
                            set length ""
                            set score ""
                            set expect ""
                            set perfection ""
                            set start -1
                            set end -1
                        }
                    }
                    unset id
                    unset count
                    unset length
                    unset score
                    unset expect
                    unset perfection
                    unset start
                    unset end
                    
                    set hNum 1
                    
                    foreach blastHit $blastHitL {
                        if {$ncbi_blast_expectCheck} {
                            if {$evalue >= [lindex $blastHit end]} {
                                set ref [ncbi_blast_get_refs [lindex $blastHit 2] $url]
                                lappend refL "'[lindex $ref 0]', '[lindex $ref 1]', '[lindex $ref 2]', '[lindex $ref 3]', '[lindex $ref 4]'"
                                if {[lindex $ref 0] ne 0} {
                                    lappend hitL "[lindex $blastHit 0] [lindex $blastHit 1] '[lindex $ref 0]', [lrange $blastHit 3 9]"
                                    incr hNum
                                }
                            }
                        } else {
                            set ref [ncbi_blast_get_refs [lindex $blastHit 2] $url]
                            lappend refL $ref
                            if {[lindex $ref 0] ne 0} {
                                lappend hitL "[lindex $blastHit 0] [lindex $blastHit 1] '[lindex $ref 0]', [lrange $blastHit 3 9]"
                                incr hNum
                            }
                        }
                    }
                    unset blastRes
                    unset blastHitL
                    if {$refL ne "" || $hitL ne ""} {
                        addRes $id_seq "$refL" "$hitL" $dbName
                    }
                    return [llength $hitL]
                }
                
                proc ncbi_blast_get_refs {key url} {
                    set giID 0
                    set prod 0
                    set org 0
                    set definition "???"
                    set organism "???"
                    set hitKey "???"
                    set product "???"
                    set refL ""
                    set complete 0
                    while {$complete eq 0} {
                        if {[catch {set token [::http::geturl "$url/sviewer/viewer.fcgi?val=$key&db=nuccore&dopt=genbank&extrafeat=976&fmt_mask=0&maxplex=1&retmode=html&log$=seqview&maxdownloadsize=1000000" -timeout 25000]}]} {
                            sleep 5000
                        } else {
                            set str [::http::data $token]
                            http::cleanup $token
                            set str [regsub -all {["]} $str ""]
                            set str [regsub -all {[><]} $str " "]
                            set str [regsub -all {[\[]} [regsub -all {[\{]} $str "("] "("]
                            set str [regsub -all {[\]]} [regsub -all {[\}]} $str ")"] "("]
                            set str "{[regsub -all {\n} $str "\} \{"]}"
                            foreach line $str {
                                if {[lindex $line 0] eq "DEFINITION"} {
                                    set definition [lrange $line 1 end]
                                } elseif {[lindex $line 0] eq "ACCESSION"} {
                                    set hitKey [lindex $line 1]
                                } elseif {[lindex $line 0] eq "VERSION"} {
                                    set giID [string range [lindex $line 2] 3 end]
                                } elseif {[lindex $line 0] eq "SOURCE" && $org ne 2} {
                                    set organism [lrange $line 1 end]
                                    set org 2
                                } elseif {[string range [lindex $line 0] 0 8] eq "/product="} {
                                    set product "[string range [lindex $line 0] 9 end] [lrange $line 1 end]"
                                    set prod 1
                                } elseif {[lindex $line 0] eq "ORIGIN" || [lindex $line 0] eq "/protein_id=" || [lindex $line 0] eq "/a"} {
                                    set prod 2
                                    break
                                } elseif {$prod eq 1} {
                                    append product " [lrange $line 1 end]"
                                } elseif {[string range [lindex $line 0] 0 12] eq "/translation="} {
                                    break
                                }
                            }
                            unset str
                            set refL "$giID $hitKey {$organism} {$definition} {$product}"
                            set complete 1
                            break
                        }
                    }
                    return $refL
                }
                proc addRes {id_seq refL hitL dbName} {
                    global phpThreadID
                    global database
                    foreach hit $hitL {
                        catch {eval [subst {thread::send $phpThreadID {::php_getSQLresults "INSERT INTO $dbName.blast_hits VALUES ($hit);"}}]}
                    }
                    unset hitL
                    
                    foreach ref $refL {
                        catch {eval [subst {thread::send $phpThreadID {::php_getSQLresults "INSERT INTO refdb.$database VALUES ($ref);"}}]}
                    }
                    unset refL
                    if {$id_seq ne "-"} {
                        catch {eval [subst {thread::send $phpThreadID {::php_getSQLresults "SELECT functions.seqinfo('$dbName', '$id_seq')"}}]}
                    }
                }
                
                proc sleep {time} {
                    global end
                    set end 0
                    after $time set end 1
                    vwait end
                }
                proc returnResult {processID status process dbName result mainThreadID mode} {
                    global env
                    global threadNum
                    global queHitNum
                    global queNoHitNum
                    global queErrorNum
                    global queTotalNum
                    
                    if {$result eq 0} {
                        incr queNoHitNum
                    } elseif {$result eq "no_respond"} {
                        incr queErrorNum
                    } elseif {$result eq "-"} {
                    } else {
                        incr queHitNum
                    }
                    
                    eval [subst {thread::send -async $mainThreadID {::ncbi_threads_finished $processID $status "$process" $dbName "$threadNum $queHitNum $queNoHitNum $queErrorNum $queTotalNum" $mode}}]
                }
                thread::wait
            }]
            if {[expr ($i + 1) * $chunkSize - 1] > [llength $id_seqL] || $i eq [expr $threadNum - 1]} {
                set idL [lrange $id_seqL [expr $i * $chunkSize] end]; eval [subst {thread::send -async $threadArrID($i) {ncbiblast $env(homeFolder) $database $program $evalue $hit \"$organism\" $url \"$idL\" [llength $id_seqL] $ncbi_blast_expectCheck $mode $processID $env(sequence_type) $env(mainThreadID) $env(phpThreadID) $i}}]
                incr ncbi_thread_num
                break
            } else {
                set idL [lrange $id_seqL [expr $i * $chunkSize] [expr ($i + 1) * $chunkSize - 1]]; eval [subst {thread::send -async $threadArrID($i) {ncbiblast $env(homeFolder) $database $program $evalue $hit \"$organism\" $url \"$idL\" [llength $id_seqL] $ncbi_blast_expectCheck $mode $processID $env(sequence_type) $env(mainThreadID) $env(phpThreadID) $i}}]
                incr ncbi_thread_num
            }
        }
        if {$mode eq "blast"} {
            for {set i 0} {$i < $ncbi_thread_num} {incr i} {
                append out "$thread_resultsArr($i)\n"
            }
            ::process_update $processID running "NCBI: 0 of [llength $id_seqL] blasted" $dbName "$out"
            unset out
        } elseif {$mode eq "ref"} {
            ::process_update $processID running "NCBI: Got 0 of [llength $id_seqL] references" $dbName 0
        }
    } else {
        if {$mode eq "blast"} {
            ::errorPopUp "Error" "No sequences selected."
        } elseif {$mode eq "ref"} {
            ::errorPopUp "Error" "No missing references."
        }
    }
}

proc ncbi_threads_finished {processID status process dbName result mode} {
    global env
    global ncbi_thread_num
    global ncbi_finished_threads
    global thread_resultsArr
    
    if {$mode eq "blast"} {
        set thread_resultsArr([lindex $result 0]) "Thread [lindex $result 0]: [expr [lindex $result 1] + [lindex $result 2] + [lindex $result 3]] of [lindex $result 4] processed; [lindex $result 1] with hits; [lindex $result 2] without hits; [lindex $result 3] with error"
        
        set out ""
        for {set i 0} {$i < $ncbi_thread_num} {incr i} {
            append out "$thread_resultsArr($i)\n"
        }
    }  
    
    if {$status eq "finished"} {
        incr ncbi_finished_threads
    } else {
        incr env($processID.ncbi_blast_num)
    }
    if {$ncbi_thread_num eq $ncbi_finished_threads} {
        if {$mode eq "ref"} {
            ::process_update $processID finished "NCBI: Got $env($processID.ncbi_blast_num) of $process references" $dbName "$ncbi_finished_threads of $ncbi_thread_num threads finished."
        } else {
            ::process_update $processID finished "NCBI: $env($processID.ncbi_blast_num) of [lindex $process 0] blasted; DB: [lindex $process 1]; program: [lindex $process 2]" $dbName "$out"
        }
        unset env($processID.ncbi_blast_num)
    } else {
        if {$mode eq "ref"} {
            ::process_update $processID running "NCBI: Got $env($processID.ncbi_blast_num) of $process references" $dbName "$result"
        } else {
            ::process_update $processID running "NCBI: $env($processID.ncbi_blast_num) of [lindex $process 0] blasted; DB: [lindex $process 1]; program: [lindex $process 2]" $dbName "$out"
        }
    }  
}
